#
# Copyright (c) 2025 CESNET z.s.p.o.
#
# This file is a part of pytest-oarepo (see https://github.com/oarepo/pytest_oarepo).
#
# pytest-oarepo is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.
#
"""UI module for pytest-oarepo."""

from __future__ import annotations



#
# Copyright (c) 2025 CESNET z.s.p.o.
#
# This file is a part of pytest-oarepo (see https://github.com/oarepo/pytest_oarepo).
#
# pytest-oarepo is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.
#
"""Test fixtures."""

from __future__ import annotations

import copy
from typing import TYPE_CHECKING, Any, Protocol

import pytest
from deepmerge import always_merger
from flask_security import login_user
from invenio_accounts.testutils import login_user_via_session
from invenio_app.factory import create_api

if TYPE_CHECKING:
    from collections.abc import Callable, Generator

    from flask import Flask
    from flask.testing import FlaskClient
    from pytest_invenio.user import UserFixtureBase
    from werkzeug.test import TestResponse


@pytest.fixture(scope="module")
def create_app(instance_path: str, entry_points: Generator[None]) -> Callable[..., Flask]:  # noqa: ARG001
    """Application factory fixture."""
    return create_api  # type: ignore [no-any-return]


@pytest.fixture
def host() -> str:
    """Return host url."""
    return "https://127.0.0.1:5000/"


@pytest.fixture
def link2testclient(host: str) -> Callable[[str, bool], str]:
    """Convert link to testclient link."""

    def _link2testclient(link: str, ui: bool = False) -> str:
        base_string = f"{host}api/" if not ui else host
        return link[len(base_string) - 1 :]

    return _link2testclient


@pytest.fixture
def default_record_json() -> dict[str, Any]:
    """Return default data for creating a record, without default workflow."""
    return {
        "metadata": {
            "creators": [
                "Creator 1",
                "Creator 2",
            ],
            "contributors": ["Contributor 1"],
            "title": "blabla",
        },
        "files": {"enabled": False},
    }


@pytest.fixture
def default_record_with_workflow_json(
    default_record_json: dict[str, Any],
) -> dict[str, Any]:
    """Return default data for creating a record."""
    return {
        **default_record_json,
        "parent": {"workflow": "default"},
    }


class PrepareRecordDataFn(Protocol):
    """Protocol for prepare_record_data fixture."""

    def __call__(
        self,
        custom_data: dict[str, Any] | None = ...,
        custom_workflow: str | None = ...,
        additional_data: dict[str, Any] | None = ...,
        add_default_workflow: bool = ...,
    ) -> dict[str, Any]:  # type: ignore[reportReturnType]
        """Call to merge input definitions into data passed to record service."""


@pytest.fixture
def prepare_record_data(default_record_json: dict[str, Any]) -> PrepareRecordDataFn:
    """Merge input definitions into data passed to record service."""

    def _merge_record_data(
        custom_data: dict[str, Any] | None = None,
        custom_workflow: str | None = None,
        additional_data: dict[str, Any] | None = None,
        add_default_workflow: bool = True,
    ) -> dict[str, Any]:
        """Merge input definitions into data passed to record service.

        :param custom_workflow: If user wants to use different workflow that the default one.
        :param additional_data: Additional data beyond the defaults that should be put into the service.
        :param add_default_workflow: Allows user to to pass data into the service without workflow -
        this might be useful for example in case of wanting to use community default workflow.
        """
        record_json = custom_data if custom_data else default_record_json
        json = copy.deepcopy(record_json)
        if add_default_workflow:
            always_merger.merge(json, {"parent": {"workflow": "default"}})
        if custom_workflow:  # specifying this assumes use of workflows
            json.setdefault("parent", {})["workflow"] = custom_workflow
        if additional_data:
            always_merger.merge(json, additional_data)

        return json

    return _merge_record_data


"""
@pytest.fixture
def vocab_cf(app: Flask, db: SQLAlchemy, cache) -> None:
    from oarepo_runtime.services.custom_fields.mappings import prepare_cf_indices

    prepare_cf_indices()
"""


class LoggedClient:
    """Logged client."""

    # TODO: - using the different clients thing?
    def __init__(self, client: FlaskClient, user_fixture: UserFixtureBase):
        """Initialize the logged client."""
        self.client: FlaskClient = client
        self.user_fixture: UserFixtureBase = user_fixture

    def _login(self) -> None:
        login_user(self.user_fixture.user, remember=True)
        login_user_via_session(self.client, email=self.user_fixture.email)

    def post(self, *args: Any, **kwargs: Any) -> TestResponse:
        """Execute POST request."""
        self._login()
        return self.client.post(*args, **kwargs)

    def get(self, *args: Any, **kwargs: Any) -> TestResponse:
        """Execute GET request."""
        self._login()
        return self.client.get(*args, **kwargs)

    def put(self, *args: Any, **kwargs: Any) -> TestResponse:
        """Execute PUT request."""
        self._login()
        return self.client.put(*args, **kwargs)

    def delete(self, *args: Any, **kwargs: Any) -> TestResponse:
        """Execute DELETE request."""
        self._login()
        return self.client.delete(*args, **kwargs)


@pytest.fixture
def logged_client(client: FlaskClient) -> Callable[[UserFixtureBase], LoggedClient]:
    """Return logged client."""

    def _logged_client(user: UserFixtureBase) -> LoggedClient:
        return LoggedClient(client, user)

    return _logged_client



#
# Copyright (c) 2025 CESNET z.s.p.o.
#
# This file is a part of pytest-oarepo (see https://github.com/oarepo/pytest_oarepo).
#
# pytest-oarepo is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.
#
"""Test classes for requests."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, cast, override

from flask_principal import UserNeed
from invenio_accounts.models import User
from invenio_requests.customizations import CommentEventType
from oarepo_runtime.services.generators import Generator
from oarepo_workflows.requests.generators import RecipientGeneratorMixin

if TYPE_CHECKING:
    from collections.abc import Collection, Mapping

    from flask_principal import Need
    from invenio_records_resources.records import Record
    from invenio_requests.customizations import RequestType


class TestEventType(CommentEventType):
    """Custom EventType."""

    type_id = "T"


class SystemUserGenerator(RecipientGeneratorMixin, Generator):
    """Generator primarily used to define system user as recipient of a request."""

    @override
    def needs(self, **kwargs: Any) -> Collection[Need]:
        return [UserNeed("system")]

    @override
    def reference_receivers(
        self,
        record: Record | None = None,
        request_type: RequestType | None = None,
        **context: Any,
    ) -> list[Mapping[str, str]]:
        return [{"user": "system"}]


class UserGenerator(RecipientGeneratorMixin, Generator):
    """Generator primarily used to define specific user as recipient of a request."""

    @override
    def __init__(self, user_email: str) -> None:
        self.user_email = user_email

    @property
    def _user_id(self) -> int:
        # id is Integer column
        return cast("int", User.query.filter_by(email=self.user_email).one().id)

    @override
    def needs(self, **kwargs: Any) -> Collection[Need]:
        return [UserNeed(self._user_id)]

    @override
    def reference_receivers(
        self,
        record: Record | None = None,
        request_type: RequestType | None = None,
        **context: Any,
    ) -> list[Mapping[str, str]]:
        return [{"user": str(self._user_id)}]


class CSLocaleUserGenerator(RecipientGeneratorMixin, Generator):
    """Generator primarily used to define specific user as recipient of a request."""

    @property
    def _user_id(self) -> int:
        users = User.query.all()
        users = [user for user in users if "locale" in user.preferences and user.preferences["locale"] == "cs"]
        if users:
            return cast("int", users[0].id)
        raise ValueError("No CS locale user found")

    @override
    def needs(self, **kwargs: Any) -> Collection[Need]:
        return [UserNeed(self._user_id)]

    @override
    def reference_receivers(
        self,
        record: Record | None = None,
        request_type: RequestType | None = None,
        **context: Any,
    ) -> list[Mapping[str, str]]:
        return [{"user": str(self._user_id)}]



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# oarepo-requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Test fixtures for requests."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Literal, Protocol

import pytest
from deepmerge import always_merger
from invenio_access.permissions import system_identity
from invenio_accounts.models import Role
from invenio_accounts.proxies import current_datastore
from invenio_records_permissions.generators import SystemProcess
from invenio_records_resources.services.errors import PermissionDeniedError
from invenio_requests.records.api import RequestEventFormat
from invenio_requests.services.generators import Receiver
from invenio_users_resources.proxies import current_groups_service

if TYPE_CHECKING:
    from collections.abc import Callable

    from flask_principal import Identity
    from invenio_db.shared import SQLAlchemy
    from invenio_records_resources.services import RecordService
    from invenio_requests.services.requests import RequestItem
    from invenio_requests.services.requests.service import RequestsService
    from pytest_invenio.user import UserFixtureBase

can_comment_only_receiver = [
    Receiver(),
    SystemProcess(),
]


class CreateRequestFn(Protocol):
    """Callable to create a request on a topic record."""

    def __call__(  # noqa PLR0913
        self,
        identity: Identity,
        id_: str,
        request_type: str,
        topic_read_method: Literal["read", "read_draft"],
        additional_data: dict[str, Any] | None = ...,
        expand: bool = ...,
        **request_kwargs: Any,
    ) -> RequestItem:  # type: ignore[reportReturnType]
        """Create request of specific type on a specific record."""


class CreateRequestOnTopicFn(Protocol):
    """Callable to create a request on a record."""

    def __call__(
        self,
        identity: Identity,
        topic_id: str,
        request_type: str,
        additional_data: dict[str, Any] | None = ...,
        expand: bool = ...,
        **request_kwargs: Any,
    ) -> RequestItem:  # type: ignore[reportReturnType]
        """Create request of specific type on a specific record."""


class SubmitRequestFn(Protocol):
    """Callable to create and submit a request on a topic record."""

    def __call__(  # noqa PLR0913
        self,
        identity: Identity,
        topic_id: str,
        request_type: str,
        create_additional_data: dict[str, Any] | None = ...,
        submit_additional_data: dict[str, Any] | None = ...,
        expand: bool = ...,
    ) -> RequestItem:  # type: ignore[reportReturnType]
        """Create and submit request of specific type on a specific record."""


def _create_role(id_: str, name: str, description: str, is_managed: bool) -> Role:
    """Create a Role/Group."""
    r = current_datastore.create_role(id=id_, name=name, description=description, is_managed=is_managed)
    current_datastore.commit()
    return r


@pytest.fixture
def role() -> Role:
    """Create a Role/Group."""
    return _create_role(
        id_="it-dep",
        name="it-dep",
        description="IT Department",
        is_managed=False,
    )


@pytest.fixture
def add_user_in_role(db: SQLAlchemy) -> Callable[[UserFixtureBase, Role | str], None]:
    """Add user to role or creates it if it doesn't exist."""

    def _add_user_in_role(user: UserFixtureBase, role_or_role_name: Role | str) -> None:
        if isinstance(role_or_role_name, str):
            try:
                role = current_groups_service.read(system_identity, role_or_role_name)._group.model.model_obj  # noqa SLF001
            except PermissionDeniedError:  # missing group in db raises this
                role = Role(name=role_or_role_name)
                db.session.add(role_or_role_name)
        else:
            role = role_or_role_name
        # TODO: UserFixtureBase is not typed
        user.user.roles.append(role)  # type: ignore[reportOptionalMemberAccess]
        db.session.commit()

    return _add_user_in_role


@pytest.fixture
def role_ui_serialization(host: str) -> dict[str, Any]:
    """UI serialization of the example role in role fixture."""
    return {
        "label": "it-dep",
        "links": {
            "avatar": f"{host}api/groups/it-dep/avatar.svg",
            "self": f"{host}api/groups/it-dep",
        },
        "reference": {"group": "it-dep"},
        "type": "group",
    }


@pytest.fixture
def events_resource_data() -> dict[str, Any]:
    """Fixture for adding required input data default values for events."""
    return {
        "payload": {
            "content": "This is an event.",
            "format": RequestEventFormat.HTML.value,
        }
    }


@pytest.fixture
def request_type_additional_data() -> dict[str, Any]:
    """Fixture for adding required input data default values specific by request type."""
    return {"publish_draft": {"payload": {"version": "1.0"}}}


@pytest.fixture
def create_request(
    request_type_additional_data: dict[str, Any],
    record_service: RecordService,
    requests_service: RequestsService,
) -> CreateRequestFn:
    """Return base function for creating a request."""

    def _create_request(  # noqa PLR0913
        identity: Identity,
        id_: str,
        request_type: str,
        topic_read_method: str,
        additional_data: dict[str, Any] | None = None,
        expand: bool = False,
        **request_kwargs: Any,
    ) -> RequestItem:
        """Create request of specific type on a specific record.

        :param identity: Identity of the caller.
        :param id_: ID of the topic record.
        :param request_type: ID of the request type.
        :param topic_read_method: Method the service uses to read the topic.
        :param additional_data: Additional data needed to create the request.
        :param expand: Expand the created request response.
        :param service_kwargs: Additional keyword arguments to pass to request service.
        """
        topic_service = record_service
        if additional_data is None:
            additional_data = {}

        if request_type in request_type_additional_data:
            additional_data = always_merger.merge(additional_data, request_type_additional_data[request_type])

        topic = getattr(topic_service, topic_read_method)(system_identity, id_)._obj  # noqa SLF001
        return requests_service.create(
            identity,
            data=additional_data,
            request_type=request_type,
            topic=topic,
            expand=expand,
            **request_kwargs,
        )

    return _create_request


@pytest.fixture
def create_request_on_draft(
    create_request: CreateRequestFn,
) -> CreateRequestOnTopicFn:
    """Fixture for creating a request on a draft."""

    def _create(
        identity: Identity,
        topic_id: str,
        request_type: str,
        additional_data: dict[str, Any] | None = None,
        expand: bool = False,
        **request_kwargs: Any,
    ) -> RequestItem:
        """Create request of specific type on a draft record.

        :param identity: Identity of the caller.
        :param topic_id: ID of the topic record.
        :param request_type: Type of the request.
        :param additional_data: Additional data needed to create the request.
        :param expand: Expand the created request response.
        :param service_kwargs: Additional keyword arguments to pass to request service.
        """
        return create_request(
            identity,
            topic_id,
            request_type,
            topic_read_method="read_draft",
            additional_data=additional_data,
            expand=expand,
            **request_kwargs,
        )

    return _create


@pytest.fixture
def create_request_on_record(
    create_request: CreateRequestFn,
) -> CreateRequestOnTopicFn:
    """Fixture for creating a request on a published record."""

    def _create(
        identity: Identity,
        topic_id: str,
        request_type: str,
        additional_data: dict[str, Any] | None = None,
        expand: bool = False,
        **request_kwargs: Any,
    ) -> RequestItem:
        """Create request of specific type on a published record.

        :param identity: Identity of the caller.
        :param topic_id: ID of the topic record.
        :param request_type: Type of the request.
        :param additional_data: Additional data needed to create the request.
        :param expand: Expand the created request response.
        :param service_kwargs: Additional keyword arguments to pass to request service.
        """
        return create_request(
            identity,
            topic_id,
            request_type,
            topic_read_method="read",
            additional_data=additional_data,
            expand=expand,
            **request_kwargs,
        )

    return _create


@pytest.fixture
def submit_request_on_draft(
    create_request_on_draft: CreateRequestOnTopicFn,
    requests_service: RequestsService,
) -> SubmitRequestFn:
    """Fixture for creating and submitting request on a draft in one call."""

    def _submit_request(  # noqa PLR0913
        identity: Identity,
        topic_id: str,
        request_type: str,
        create_additional_data: dict[str, Any] | None = None,
        submit_additional_data: dict[str, Any] | None = None,
        expand: bool = False,
    ) -> RequestItem:
        """Create and submit request of specific type on a specific record.

        :param identity: Identity of tha caller.
        :param topic_id: ID of the topic record.
        :param request_type: Type of the request.
        :param create_additional_data: Additional data needed to create the request.
        :param submit_additional_data: Additional data passed to the submit action.
        :param expand: Expand the created request response.
        :param service_kwargs: Additional keyword arguments to pass to request service.
        """
        create_response = create_request_on_draft(
            identity, topic_id, request_type, additional_data=create_additional_data
        )
        return requests_service.execute_action(
            identity,
            id_=create_response["id"],
            action="submit",
            data=submit_additional_data,
            expand=expand,
        )

    return _submit_request


@pytest.fixture
def submit_request_on_record(
    create_request_on_record: CreateRequestOnTopicFn,
    requests_service: RequestsService,
) -> SubmitRequestFn:
    """Fixture for creating and submitting request on a published record in one call."""

    def _submit_request(  # noqa PLR0913
        identity: Identity,
        topic_id: str,
        request_type: str,
        create_additional_data: dict[str, Any] | None = None,
        submit_additional_data: dict[str, Any] | None = None,
        expand: bool = False,
    ) -> RequestItem:
        """Create and submit request of specific type on a specific record.

        :param client: Client instance.
        :param record: Record on which the request should be created.
        :param request_type: Which type of request should be created.
        :param create_additional_data: Additional data needed to create the request.
        :param submit_additional_data: Additional data passed to the submit action.
        :param expand: Expand the submitted request response.
        """
        create_response = create_request_on_record(
            identity, topic_id, request_type, additional_data=create_additional_data
        )
        return requests_service.execute_action(
            identity,
            id_=create_response["id"],
            action="submit",
            data=submit_additional_data,
            expand=expand,
        )

    return _submit_request



#
# Copyright (c) 2025 CESNET z.s.p.o.
#
# This file is a part of pytest-oarepo (see https://github.com/oarepo/pytest_oarepo).
#
# pytest-oarepo is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.
#
"""Fixtures for creating test records."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Protocol

import pytest
from invenio_access.permissions import system_identity

if TYPE_CHECKING:
    from flask_principal import Identity
    from invenio_drafts_resources.services import RecordService

    from pytest_oarepo.files import UploadFileFn
    from pytest_oarepo.fixtures import PrepareRecordDataFn


class CreateRecordFn(Protocol):
    """Callable to create a record."""

    def __call__(
        self,
        identity: Identity,
        custom_data: dict[str, Any] | None = ...,
        additional_data: dict[str, Any] | None = ...,
        custom_workflow: str | None = ...,
        expand: bool = ...,
        **service_kwargs: Any,
    ) -> dict[str, Any]:  # type: ignore[reportReturnType]
        """Create instance of a record."""


class CreateRecordWithFilesFn(Protocol):
    """Callable to create a record with a file."""

    def __call__(  # noqa PLR0913
        self,
        identity: Identity,
        custom_data: dict[str, Any] | None = ...,
        additional_data: dict[str, Any] | None = ...,
        custom_workflow: str | None = ...,
        expand: bool = ...,
        file_name: str = ...,
        custom_file_metadata: dict[str, Any] | None = ...,
        **service_kwargs: Any,
    ) -> dict[str, Any]:  # type: ignore[reportReturnType]
        """Create instance of a record with a file."""


@pytest.fixture
def draft_factory(record_service: RecordService, prepare_record_data: PrepareRecordDataFn) -> CreateRecordFn:
    """Call to instance a draft."""

    def draft(
        identity: Identity,
        custom_data: dict[str, Any] | None = None,
        additional_data: dict[str, Any] | None = None,
        custom_workflow: str | None = None,
        expand: bool = False,
        **service_kwargs: Any,
    ) -> dict[str, Any]:  # type: ignore[reportReturnType]
        """Create instance of a draft.

        :param identity: Identity of the caller.
        :param custom_data: If defined, the default record data are overwritten.
        :param additional_data: If defined, the additional data are merged with the default data.
        :param custom_workflow: Define to use custom workflow.
        :param expand: Expand the response.
        :param service_kwargs: Additional keyword arguments to pass to the service.
        """
        # TODO: possibly support for more model types?
        # like this perhaps
        # ruff ok service = record_service(model) if isinstance(record_service, callable) else record_service

        json = prepare_record_data(custom_data, custom_workflow, additional_data)
        draft = record_service.create(identity=identity, data=json, expand=expand, **service_kwargs)
        # TODO: to_dict() is typed as dict[str, Any] in RecordItem, idk why it complains here
        # unified interface
        return draft.to_dict()  # type: ignore[no-any-return]

    return draft


@pytest.fixture
def record_factory(record_service: RecordService, draft_factory: CreateRecordFn) -> CreateRecordFn:
    """Call to instance a published record."""

    def record(
        identity: Identity,
        custom_data: dict[str, Any] | None = None,
        additional_data: dict[str, Any] | None = None,
        custom_workflow: str | None = None,
        expand: bool = False,
        **service_kwargs: Any,
    ) -> dict[str, Any]:
        """Create instance of a published record.

        :param identity: Identity of the caller.
        :param custom_data: If defined, the default record data are overwritten.
        :param additional_data: If defined, the additional data are merged with the default data.
        :param custom_workflow: Define to use custom workflow.
        :param expand: Expand the response.
        :param service_kwargs: Additional keyword arguments to pass to the service.
        """
        draft = draft_factory(
            identity,
            custom_data=custom_data,
            additional_data=additional_data,
            custom_workflow=custom_workflow,
            **service_kwargs,
        )
        record = record_service.publish(system_identity, draft["id"], expand=expand)
        return record.to_dict()  # type: ignore[no-any-return]

    return record


@pytest.fixture
def record_with_files_factory(
    record_service: RecordService,
    draft_factory: CreateRecordFn,
    default_record_with_workflow_json: dict[str, Any],
    upload_file: UploadFileFn,
) -> CreateRecordWithFilesFn:
    """Call to instance a published record with a file."""

    def record(  # noqa PLR0913
        identity: Identity,
        custom_data: dict[str, Any] | None = None,
        additional_data: dict[str, Any] | None = None,
        custom_workflow: str | None = None,
        expand: bool = False,
        # kept for API parity
        file_name: str = "test.pdf",  # noqa ARG001
        custom_file_metadata: dict[str, Any] | None = None,  # noqa ARG001
        **service_kwargs: Any,
    ) -> dict[str, Any]:
        """Create instance of a published record.

        :param identity: Identity of tha caller.
        :param custom_data: If defined, the default record data are overwritten.
        :param additional_data: If defined, the additional data are merged with the default data.
        :param custom_workflow: Define to use custom workflow.
        :param expand: Expand the response.
        :param file_name: Name of the file to upload.
        :param custom_file_metadata: Define to use custom file metadata.
        :param service_kwargs: Additional keyword arguments to pass to the service.
        """
        if "files" in default_record_with_workflow_json and "enabled" in default_record_with_workflow_json["files"]:
            if not additional_data:
                additional_data = {}
            additional_data.setdefault("files", {}).setdefault("enabled", True)
        draft = draft_factory(
            identity,
            custom_data=custom_data,
            additional_data=additional_data,
            custom_workflow=custom_workflow,
            **service_kwargs,
        )
        # _draft_files is not typed
        files_service = record_service._draft_files  # type: ignore[reportAttributeAccessIssue]  # noqa: SLF001
        upload_file(identity, draft["id"], files_service)
        record = record_service.publish(
            system_identity,
            draft["id"],
            expand=expand,
        )
        return record.to_dict()  # type: ignore[no-any-return]

    return record



#
# Copyright (c) 2025 CESNET z.s.p.o.
#
# This file is a part of pytest-oarepo (see https://github.com/oarepo/pytest-oarepo).
#
# pytest-oarepo is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.
#
"""Base module for pytest-oarepo.

This module provides testing utilities and fixtures in OARepo applications.
"""

from __future__ import annotations

__version__ = "2.0.2"

__all__ = [
    "__version__",
]



#
# Copyright (c) 2025 CESNET z.s.p.o.
#
# This file is a part of pytest-oarepo (see https://github.com/oarepo/pytest_oarepo).
#
# pytest-oarepo is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.
#
"""Test fixtures for working with files."""

from __future__ import annotations

from io import BytesIO
from typing import TYPE_CHECKING, Any, Protocol, cast

import pytest

if TYPE_CHECKING:
    from flask_principal import Identity
    from invenio_records_resources.services.files.results import FileItem
    from invenio_records_resources.services.files.service import FileService


class UploadFileFn(Protocol):
    """Callable to upload a file to a record."""

    def __call__(
        self,
        identity: Identity,
        record_id: str,
        files_service: FileService,
        file_name: str = ...,
        custom_file_metadata: dict[str, Any] | None = ...,
    ) -> FileItem:  # type: ignore[reportReturnType]
        """Upload a file to a record."""


@pytest.fixture
def file_metadata() -> dict[str, Any]:
    """Return default file metadata."""
    return {"title": "Test file"}


@pytest.fixture
def upload_file(file_metadata: dict[str, Any]) -> UploadFileFn:
    """Fixture to upload a file to a record."""

    def _upload_file(
        identity: Identity,
        record_id: str,
        files_service: FileService,
        file_name: str = "test.pdf",
        custom_file_metadata: dict[str, Any] | None = None,
    ) -> FileItem:
        """Upload a default file to a record.

        :param identity: Identity of the requester.
        :param record_id: Id of the record to be uploaded on.
        :param files_service: Service to upload the file.
        :param file_name: Name of the file to be uploaded.
        :param custom_file_metadata: Custom metadata to be uploaded.
        """
        actual_file_metadata = custom_file_metadata if custom_file_metadata else file_metadata
        files_service.init_files(
            identity,
            record_id,
            data=[
                {"key": file_name, "metadata": actual_file_metadata},
            ],
        )
        files_service.set_file_content(identity, record_id, file_name, stream=BytesIO(b"testfile"))
        return cast("FileItem", files_service.commit_file(identity, record_id, file_name))

    return _upload_file



#
# Copyright (c) 2025 CESNET z.s.p.o.
#
# This file is a part of pytest-oarepo (see https://github.com/oarepo/pytest_oarepo).
#
# pytest-oarepo is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.
#
"""Test functions for requests."""

from __future__ import annotations

from typing import Any, cast


def get_request_type(request_types_json: list[dict[str, Any]], request_type: str) -> dict[str, Any] | None:
    """Get request type from request types json."""
    selected_entry = [entry for entry in request_types_json if entry["type_id"] == request_type]
    if not selected_entry:
        return None
    return selected_entry[0]


def get_request_create_link(request_types_json: list[dict[str, Any]], request_type: str) -> str:
    """Get request create link from request types json."""
    selected_entry = get_request_type(request_types_json, request_type)
    if not selected_entry:
        raise ValueError(f"Request type {request_type} not found in request types")
    return cast("str", selected_entry["links"]["actions"]["create"])



#
# Copyright (c) 2025 CESNET z.s.p.o.
#
# This file is a part of pytest-oarepo (see https://github.com/oarepo/pytest_oarepo).
#
# pytest-oarepo is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.
#
"""UI classes for pytest-oarepo."""

from __future__ import annotations

from typing import override

from flask_webpackext.manifest import (
    JinjaManifest,
    JinjaManifestEntry,
    JinjaManifestLoader,
)


#
# Mock the webpack manifest to avoid having to compile the full assets.
#
class MockJinjaManifest(JinjaManifest):
    """Mock manifest."""

    def __getitem__(self, key: str) -> JinjaManifestEntry:
        """Get a manifest entry."""
        return JinjaManifestEntry(key, [key])

    def __getattr__(self, name: str) -> JinjaManifestEntry:
        """Get a manifest entry."""
        return JinjaManifestEntry(name, [name])


class MockManifestLoader(JinjaManifestLoader):
    """Manifest loader creating a mocked manifest."""

    @override
    def load(self, filepath: str) -> MockJinjaManifest:
        """Load the manifest."""
        return MockJinjaManifest()



#
# Copyright (c) 2025 CESNET z.s.p.o.
#
# This file is a part of pytest-oarepo (see https://github.com/oarepo/pytest_oarepo).
#
# pytest-oarepo is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.
#
"""Vocabularies module for pytest-oarepo.

This module provides testing utilities and fixtures for working with vocabularies
in OARepo applications.
"""

from __future__ import annotations

"""
from pytest_oarepo.vocabularies.loading import test_vocabularies

__all__ = ("test_vocabularies",)
"""



#
# Copyright (c) 2025 CESNET z.s.p.o.
#
# This file is a part of pytest-oarepo (see https://github.com/oarepo/pytest_oarepo).
#
# pytest-oarepo is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.
#
"""Vocabularies config module for pytest-oarepo."""

from __future__ import annotations

# TODO: this should have been chnged in new RDM?
"""
class FineGrainedPermissionPolicy(EveryonePermissionPolicy):
    can_create_removalreasons = (SystemProcess(), AnyUser()]
    can_update_removalreasons = [
        SystemProcess(),
        NonDangerousVocabularyOperation(AnyUser()),
    ]
    can_delete_removalreasons = [SystemProcess(), AnyUser()]


VOCABULARIES_TEST_CONFIG = {
    "VOCABULARIES_PERMISSIONS_PRESETS": ["fine-grained"],
    "OAREPO_PERMISSIONS_PRESETS": {"fine-grained": FineGrainedPermissionPolicy},
    "VOCABULARIES_SERVICE_CONFIG": VocabulariesConfig,
}
"""



#
# Copyright (c) 2025 CESNET z.s.p.o.
#
# This file is a part of pytest-oarepo (see https://github.com/oarepo/pytest_oarepo).
#
# pytest-oarepo is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.
#
"""Module for loading vocabulary fixtures."""

from __future__ import annotations

# TODO: datastreams scrapped
"""
from pathlib import Path
from typing import TYPE_CHECKING

import pytest
from invenio_vocabularies.records.api import Vocabulary
from oarepo_runtime.datastreams.fixtures import FixturesCallback, load_fixtures

if TYPE_CHECKING:
    from oarepo_runtime.datastreams import StreamBatch


@pytest.fixture
def test_vocabularies() -> None:


    class ErrCallback(FixturesCallback):

        def batch_finished(self, batch: StreamBatch) -> None:
            if batch.failed_entries:
                pass
            super().batch_finished(batch)

    callback = ErrCallback()
    load_fixtures(Path(__file__).parent / "data", callback=callback, system_fixtures=False)
    Vocabulary.index.refresh()
"""



#
# Copyright (c) 2025 CESNET z.s.p.o.
#
# This file is a part of pytest-oarepo (see https://github.com/oarepo/pytest_oarepo).
#
# pytest-oarepo is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.
#
"""Fixtures for creating test users."""

from __future__ import annotations

import base64
import os
from typing import TYPE_CHECKING

import pytest
from sqlalchemy.exc import IntegrityError

from pytest_oarepo.functions import _index_users

if TYPE_CHECKING:
    from flask import Flask
    from invenio_db.shared import SQLAlchemy
    from pytest_invenio.user import UserFixtureBase


@pytest.fixture
def password() -> str:
    """Return password."""
    return base64.b64encode(os.urandom(16)).decode("utf-8")


def _create_user(user_fixture: UserFixtureBase, app: Flask, db: SQLAlchemy) -> None:
    """Create users, reusing it if it already exists."""
    try:
        user_fixture.create(app, db)
    except IntegrityError:
        datastore = app.extensions["security"].datastore
        user_fixture._user = datastore.get_user_by_email(  # noqa: SLF001
            user_fixture.email
        )
        user_fixture._app = app  # noqa: SLF001


@pytest.fixture
def users(app: Flask, db: SQLAlchemy, UserFixture: type[UserFixtureBase], password: str) -> list[UserFixtureBase]:  # noqa N803
    """Predefined user fixtures."""
    user1 = UserFixture(
        email="user1@example.org",
        password=password,
        active=True,
        confirmed=True,
        user_profile={
            "affiliations": "CERN",
        },
        preferences={"locale": "en", "visibility": "public"},
    )
    _create_user(user1, app, db)

    user2 = UserFixture(
        email="user2@example.org",
        password=password,
        username="beetlesmasher",
        active=True,
        confirmed=True,
        user_profile={
            "affiliations": "CERN",
        },
        preferences={"locale": "en", "visibility": "public"},
    )
    _create_user(user2, app, db)

    user3 = UserFixture(
        email="user3@example.org",
        password=password,
        username="beetlesmasherXXL",
        user_profile={
            "full_name": "Maxipes Fik",
            "affiliations": "CERN",
        },
        active=True,
        confirmed=True,
        preferences={"locale": "en", "visibility": "public"},
    )
    _create_user(user3, app, db)

    user4 = UserFixture(
        email="user4@example.org",
        password=password,
        username="african",
        preferences={
            "timezone": "Africa/Dakar",  # something without daylight saving time; +0.0
            "locale": "en",
            "visibility": "public",
        },
        user_profile={
            "affiliations": "CERN",
        },
        active=True,
        confirmed=True,
    )
    _create_user(user4, app, db)

    user5 = UserFixture(
        email="user5@example.org",
        password=password,
        username="mexican",
        preferences={
            "timezone": "America/Mexico_City",  # something without daylight saving time
            "locale": "en",
            "visibility": "public",
        },
        user_profile={
            "affiliations": "CERN",
        },
        active=True,
        confirmed=True,
    )
    _create_user(user5, app, db)

    db.session.commit()
    _index_users()
    return [user1, user2, user3, user4, user5]


@pytest.fixture
def user_with_cs_locale(
    app: Flask,
    db: SQLAlchemy,
    UserFixture: type[UserFixtureBase],  # noqa N803
    password: str,
) -> UserFixtureBase:
    """User with Czech locale."""
    u = UserFixture(
        email="pat@mat.cz",
        password=password,
        username="patmat",
        user_profile={
            "full_name": "patmat",
            "affiliations": "cesnet",
        },
        preferences={"locale": "cs"},
        active=True,
        confirmed=True,
    )
    _create_user(u, app, db)
    db.session.commit()
    _index_users()
    return u



#
# Copyright (c) 2025 CESNET z.s.p.o.
#
# This file is a part of pytest-oarepo (see https://github.com/oarepo/pytest_oarepo).
#
# pytest-oarepo is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.
#
"""Util used in testing."""

from __future__ import annotations

from collections import defaultdict

from flask import g
from invenio_users_resources.proxies import current_users_service


def _dict_diff(dict1: dict, dict2: dict, path: str = "") -> dict[str, list[str]]:
    ret = defaultdict(list)
    for key in dict1:  # noqa PLC0206
        # Construct path to current element
        new_path = key if path == "" else f"{path}.{key}"

        # Check if the key is in the second dictionary
        if key not in dict2:
            ret["second dict missing"].append(f"{new_path}: Key missing in the second dictionary")
            continue

        # If both values are dictionaries, do a recursive call
        if isinstance(dict1[key], dict) and isinstance(dict2[key], dict):
            sub_result = _dict_diff(dict1[key], dict2[key], new_path)
            ret.update(sub_result)
        # Check if values are the same
        elif dict1[key] != dict2[key]:
            ret["different values"].append(f"{new_path}: {dict1[key]} != {dict2[key]}")

    # Check for keys in the second dictionary but not in the first
    for key in dict2:
        if key not in dict1:
            new_path = key if path == "" else f"{path}.{key}"
            ret["first dict missing"].append(f"{new_path}: Key missing in the first dictionary")
    return ret


# TODO: scrap?
def is_valid_subdict(subdict: dict, dict_: dict) -> bool:
    """Check if subdict is a valid subdict of dict."""
    diff = _dict_diff(subdict, dict_)
    return "different values" not in diff and "second dict missing" not in diff


def _index_users() -> None:
    """Index users."""
    current_users_service.indexer.process_bulk_queue()
    current_users_service.indexer.refresh()


# TODO: scrap?
def clear_babel_context() -> None:
    """Clear babel context."""
    # for invenio 12
    try:
        from flask_babel import SimpleNamespace
    except ImportError:
        return
    g._flask_babel = SimpleNamespace()  # noqa: SLF001



#
# Copyright (c) 2025 CESNET z.s.p.o.
#
# This file is a part of pytest-oarepo (see https://github.com/oarepo/pytest_oarepo).
#
# pytest-oarepo is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.
#
"""UI fixtures for pytest-oarepo."""

from __future__ import annotations

from typing import Any

import pytest

from .classes import MockManifestLoader


@pytest.fixture(scope="module")
def app_config(app_config: dict[str, Any]) -> dict[str, Any]:
    """Override pytest-invenio app_config fixture with custom test settings."""
    # Misc / frontend
    app_config.update(
        IIIF_FORMATS=["jpg", "png"],
        APP_RDM_RECORD_THUMBNAIL_SIZES=[500],
        WEBPACKEXT_MANIFEST_LOADER=MockManifestLoader,
    )

    return app_config
